# Copyright 2021 Ilango Rajagopal
# Licensed under GPL-3.0-only

"""This module provides utilities to read the grm grammar format"""

from pypargen.grm.parser import GrmParser

__all__ = ["GrmParser"]
