#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pypargen', 'pypargen.base', 'pypargen.grm', 'pypargen.lexer', 'pypargen.lr1']

package_data = \
{'': ['*']}

extras_require = \
{'test': ['pytest', 'pytest-cov']}

setup(name='pypargen',
      version='0.3.0',
      description='A parser generator built with Python',
      author='Ilango Rajagopal',
      author_email='ilangokul@gmail.com',
      url='https://github.com/ilango100/pypargen',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='>=3.9',
     )
